package dev.fyloz.colorrecipesexplorer.service.config

import dev.fyloz.colorrecipesexplorer.SUPPORTED_DATABASE_VERSION
import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import dev.fyloz.colorrecipesexplorer.emergencyMode
import dev.fyloz.colorrecipesexplorer.model.CannotSetComputedConfigurationException
import dev.fyloz.colorrecipesexplorer.model.Configuration
import dev.fyloz.colorrecipesexplorer.model.ConfigurationType
import dev.fyloz.colorrecipesexplorer.model.configuration
import dev.fyloz.colorrecipesexplorer.repository.ConfigurationRepository
import dev.fyloz.colorrecipesexplorer.service.create
import dev.fyloz.colorrecipesexplorer.utils.excludeAll
import org.slf4j.Logger
import org.springframework.boot.info.BuildProperties
import org.springframework.context.annotation.Lazy
import org.springframework.data.repository.findByIdOrNull
import org.springframework.stereotype.Component
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream
import java.time.LocalDate
import java.time.LocalDateTime
import java.time.ZoneId
import java.util.*

const val CONFIGURATION_FILE_PATH = "config.properties"
const val CONFIGURATION_FILE_COMMENT = "---Color Recipes Explorer configuration---"

interface ConfigurationSource {
    fun get(type: ConfigurationType): Configuration?
    fun set(configuration: Configuration)
    fun set(configurations: Iterable<Configuration>)
}

@Component("configurationSource")
class CompositeConfigurationSource(
    @Lazy private val configurationRepository: ConfigurationRepository,
    private val properties: CreProperties,
    private val buildInfo: BuildProperties,
    private val logger: Logger
) : ConfigurationSource {
    private val repository by lazy { RepositoryConfigurationSource(configurationRepository) }
    private val file by lazy {
        FileConfigurationSource("${properties.configDirectory}/$CONFIGURATION_FILE_PATH")
    }
    private val computed by lazy {
        ComputedConfigurationSource(buildInfo)
    }

    override fun get(type: ConfigurationType) =
        when {
            type.file -> file.get(type)
            type.computed -> computed.get(type)
            !emergencyMode -> repository.get(type)
            else -> null
        }

    override fun set(configuration: Configuration) =
        when {
            configuration.type.file -> file.set(configuration)
            configuration.type.computed -> throw CannotSetComputedConfigurationException(configuration.type)
            !emergencyMode -> repository.set(configuration)
            else -> {
            }
        }

    override fun set(configurations: Iterable<Configuration>) {
        val mutableConfigurations = configurations.toMutableList()
        val fileConfigurations = mutableConfigurations.excludeAll { it.type.file }
        val repositoryConfigurations = mutableConfigurations.excludeAll { !emergencyMode }

        repository.set(repositoryConfigurations)
        file.set(fileConfigurations)

        mutableConfigurations.forEach {
            logger.warn("Could not find where to store updated value of configuration '${it.key}'")
        }
    }
}

private class RepositoryConfigurationSource(
    private val repository: ConfigurationRepository
) : ConfigurationSource {
    override fun get(type: ConfigurationType) =
        repository.findByIdOrNull(type.key)?.toConfiguration()

    override fun set(configuration: Configuration) {
        repository.save(configuration.toEntity())
    }

    override fun set(configurations: Iterable<Configuration>) =
        configurations.forEach { set(it) }
}

private class FileConfigurationSource(
    private val configFilePath: String
) : ConfigurationSource {
    private val properties = Properties().apply {
        with(File(configFilePath)) {
            if (!this.exists()) this.create()
            FileInputStream(this).use {
                this@apply.load(it)
            }
        }
    }

    override fun get(type: ConfigurationType) =
        if (properties.containsKey(type.key))
            configuration(
                type,
                getConfigurationContent(type.key),
                LocalDateTime.parse(getConfigurationContent(configurationLastUpdateKey(type.key)))
            )
        else null

    override fun set(configuration: Configuration) {
        setConfigurationContent(configuration.type.key, configuration.content)
        save()
    }

    override fun set(configurations: Iterable<Configuration>) {
        configurations.forEach {
            setConfigurationContent(it.type.key, it.content)
        }
        save()
    }

    fun save() {
        FileOutputStream(configFilePath).use {
            properties.store(it, CONFIGURATION_FILE_COMMENT)
        }
    }

    private fun getConfigurationContent(key: String) =
        properties[key] as String

    private fun setConfigurationContent(key: String, content: String) {
        properties[key] = content
        properties[configurationLastUpdateKey(key)] = LocalDateTime.now().toString()
    }

    private fun configurationLastUpdateKey(key: String) = "$key.last-updated"
}

private class ComputedConfigurationSource(
    private val buildInfo: BuildProperties
) : ConfigurationSource {
    override fun get(type: ConfigurationType) = configuration(
        type, when (type) {
            ConfigurationType.EMERGENCY_MODE_ENABLED -> emergencyMode
            ConfigurationType.BUILD_VERSION -> buildInfo.version
            ConfigurationType.BUILD_TIME -> LocalDate.ofInstant(buildInfo.time, ZoneId.systemDefault()).toString()
            ConfigurationType.DATABASE_SUPPORTED_VERSION -> SUPPORTED_DATABASE_VERSION
            ConfigurationType.JAVA_VERSION -> Runtime.version()
            ConfigurationType.OPERATING_SYSTEM -> "${System.getProperty("os.name")} ${System.getProperty("os.version")} ${
                System.getProperty(
                    "os.arch"
                )
            }"
            else -> throw IllegalArgumentException("Cannot get the value of the configuration with the key ${type.key} because it is not a computed configuration")
        }.toString()
    )

    override fun set(configuration: Configuration) {
        throw UnsupportedOperationException("Cannot set computed configurations")
    }

    override fun set(configurations: Iterable<Configuration>) {
        throw UnsupportedOperationException("Cannot set computed configurations")
    }
}
